//AY-3-8910 sound chip emulator running on a given sample rate (with aliasing and low quality)
//shiru8bit 02'25

enum {
	AY_REG_TONE_A_LSB = 0,
	AY_REG_TONE_A_MSB,
	AY_REG_TONE_B_LSB,
	AY_REG_TONE_B_MSB,
	AY_REG_TONE_C_LSB,
	AY_REG_TONE_C_MSB,
	AY_REG_NOISE,
	AY_REG_MIXER,
	AY_REG_AMP_A,
	AY_REG_AMP_B,
	AY_REG_AMP_C,
	AY_REG_ENV_LSB,
	AY_REG_ENV_MSB,
	AY_REG_ENV_MODE,
	AY_REG_PORT_A,
	AY_REG_PORT_B
};

#define AY_PRELIMINARY_DIVIDER	8

#define AY_ENV_FLAGS_HOLD		0x01
#define AY_ENV_FLAGS_ALTERNATE	0x02
#define AY_ENV_FLAGS_ATTACK		0x04
#define AY_ENV_FLAGS_CONTINUE	0x08

struct {

	uint32_t tone_a_cnt;	//both counter and dividers are 16:8
	uint32_t tone_a_div;	//R0,R1 in 16:8 format; it is never 0, 0 considered 1<<8
	uint32_t tone_b_cnt;
	uint32_t tone_b_div;	//R2,R3 in 16:8 format
	uint32_t tone_c_cnt;
	uint32_t tone_c_div;	//R4,R5 in 16:8 format

	uint8_t tone_a_out;	//actual tone output is in bit 0
	uint8_t tone_b_out;
	uint8_t tone_c_out;

	uint16_t noise_cnt;	//8:8
	uint16_t noise_div;
	uint16_t noise_lfsr;
	uint8_t noise_out;	//actual noise output is in bit 0

	uint8_t mixer;		//R7

	uint8_t amp_a;		//R8
	uint8_t amp_b;		//R9
	uint8_t amp_c;		//R10

	uint32_t env_cnt;		//16:1:8 format, to provide extra divide by 2
	uint32_t env_div;
	uint8_t env_mode;		//R13
	int8_t env_amp;		//envelope amplitude counter
	int8_t env_amp_delta;	//count direction, -1, 1, or 0 to stop

	uint8_t dac_a;		//DACs outputs
	uint8_t dac_b;
	uint8_t dac_c;

	int16_t frac_clock;	//16:8 step for tone counters based on sample rate

} AYE;



#ifndef VOLUME_TABLE_DIVIDER
#define VOLUME_TABLE_DIVIDER	1
#endif

#if VOLUME_TABLE_DIVIDER >= 256
	const uint8_t AY_EmuVolumeTable[16] = {
#else
const uint16_t AY_EmuVolumeTable[16] = {
#endif
	0 / VOLUME_TABLE_DIVIDER,
	836 / VOLUME_TABLE_DIVIDER,
	1212 / VOLUME_TABLE_DIVIDER,
	1773 / VOLUME_TABLE_DIVIDER,
	2619 / VOLUME_TABLE_DIVIDER,
	3875 / VOLUME_TABLE_DIVIDER,
	5397 / VOLUME_TABLE_DIVIDER,
	8823 / VOLUME_TABLE_DIVIDER,
	10392 / VOLUME_TABLE_DIVIDER,
	16706 / VOLUME_TABLE_DIVIDER,
	23339 / VOLUME_TABLE_DIVIDER,
	29292 / VOLUME_TABLE_DIVIDER,
	36969 / VOLUME_TABLE_DIVIDER,
	46421 / VOLUME_TABLE_DIVIDER,
	55195 / VOLUME_TABLE_DIVIDER,
	65535 / VOLUME_TABLE_DIVIDER
};



void AY_EmuInit(uint32_t chip_clock, uint32_t sample_rate)
{
	AYE.frac_clock = chip_clock / AY_PRELIMINARY_DIVIDER * 256 / sample_rate;
}



void AY_EmuReset(void)
{
	AYE.tone_a_cnt = 0;
	AYE.tone_b_cnt = 0;
	AYE.tone_c_cnt = 0;
	AYE.tone_a_div = 1 << 8;
	AYE.tone_b_div = 1 << 8;
	AYE.tone_c_div = 1 << 8;
	AYE.tone_a_out = 0;
	AYE.tone_b_out = 0;
	AYE.tone_c_out = 0;

	AYE.noise_cnt = 0;
	AYE.noise_lfsr = 0xffff; 
	AYE.noise_div = 1 << 8;
	AYE.noise_out = 0;
	
	AYE.mixer = 0xff;

	AYE.amp_a = 0;
	AYE.amp_b = 0;
	AYE.amp_c = 0;

	AYE.env_cnt = 0;
	AYE.env_div = 1 << 12;
	AYE.env_amp = 0;
	AYE.env_amp_delta = 0;
	AYE.env_mode = 0;

	AYE.dac_a = 0;
	AYE.dac_b = 0;
	AYE.dac_c = 0;
}



void AY_EmuWriteReg(uint8_t reg, uint8_t value)
{
	switch (reg)
	{
	case AY_REG_TONE_A_LSB:
		AYE.tone_a_div = (AYE.tone_a_div & 0xff00ff) | ((uint32_t)value << 8);
		break;

	case AY_REG_TONE_A_MSB:
		AYE.tone_a_div = (AYE.tone_a_div & 0x00ffff) | ((uint32_t)(value & 0x0f) << 16);
		break;

	case AY_REG_TONE_B_LSB:
		AYE.tone_b_div = (AYE.tone_b_div & 0xff00ff) | ((uint32_t)value << 8);
		break;

	case AY_REG_TONE_B_MSB:
		AYE.tone_b_div = (AYE.tone_b_div & 0x00ffff) | ((uint32_t)(value & 0x0f) << 16);
		break;

	case AY_REG_TONE_C_LSB:
		AYE.tone_c_div = (AYE.tone_c_div & 0xff00ff) | ((uint32_t)value << 8);
		break;

	case AY_REG_TONE_C_MSB:
		AYE.tone_c_div = (AYE.tone_c_div & 0x00ffff) | ((uint32_t)(value & 0x0f) << 16);
		break;

	case AY_REG_NOISE:
		AYE.noise_div = (AYE.noise_div & 0x00ff) | ((value & 0x3f) << 8);
		if (AYE.noise_div < 0x100) AYE.noise_div = 0x100;
		break;

	case AY_REG_MIXER:
		AYE.mixer = value & 0x3f;
		break;

	case AY_REG_AMP_A:
		AYE.amp_a = value & 0x1f;
		break;

	case AY_REG_AMP_B:
		AYE.amp_b = value & 0x1f;
		break;

	case AY_REG_AMP_C:
		AYE.amp_c = value & 0x1f;
		break;

	case AY_REG_ENV_LSB:
		AYE.env_div = (AYE.env_div & 0xfe01ff) | ((uint32_t)value << 9);
		break;

	case AY_REG_ENV_MSB:
		AYE.env_div = (AYE.env_div & 0x01ffff) | ((uint32_t)value << 17);
		break;

	case AY_REG_ENV_MODE:
		AYE.env_mode = value;
		AYE.env_cnt = 0;
		if (AYE.env_mode & AY_ENV_FLAGS_ATTACK)
		{
			AYE.env_amp = 0;
			AYE.env_amp_delta = 1;
		}
		else
		{
			AYE.env_amp = 15;
			AYE.env_amp_delta = -1;
		}
		break;
	}

	if (AYE.tone_a_div < 0x100) AYE.tone_a_div = 0x100;
	if (AYE.tone_b_div < 0x100) AYE.tone_b_div = 0x100;
	if (AYE.tone_c_div < 0x100) AYE.tone_c_div = 0x100;

	if (AYE.env_div < 0x200) AYE.env_div = 0x100;	//envelope half-rate for divider 0
}



inline void AY_EmuAdvanceSample(void)
{
	//tone generator A

	AYE.tone_a_cnt += AYE.frac_clock;

	while (AYE.tone_a_cnt >= AYE.tone_a_div)	//if sample rate is low, counter overflow may occur a few times per sample
	{
		++AYE.tone_a_out;	//so output state will flip as many times

		AYE.tone_a_cnt -= AYE.tone_a_div;
	}

	//tone generator B

	AYE.tone_b_cnt += AYE.frac_clock;

	while (AYE.tone_b_cnt >= AYE.tone_b_div)
	{
		++AYE.tone_b_out;

		AYE.tone_b_cnt -= AYE.tone_b_div;
	}

	//tone generator C

	AYE.tone_c_cnt += AYE.frac_clock;

	while (AYE.tone_c_cnt >= AYE.tone_c_div)
	{
		++AYE.tone_c_out;

		AYE.tone_c_cnt -= AYE.tone_c_div;
	}

	//noise generator

	AYE.noise_cnt += AYE.frac_clock;

	if (AYE.noise_cnt >= AYE.noise_div)
	{
		while (AYE.noise_cnt >= AYE.noise_div) AYE.noise_cnt -= AYE.noise_div;

		//there is no reason to generate more noise bits even if overflow occured a few times
		//it won't make audible difference, because its spectrum limited by the sample rate anyways

		unsigned char di = (AYE.noise_out ^ ((AYE.noise_lfsr >> 13) & 1)) ^ 1;

		AYE.noise_out = (AYE.noise_lfsr & 0x8000) ? 1 : 0;
		AYE.noise_lfsr = (AYE.noise_lfsr << 1) | di;
	}

	//envelope generator

	AYE.env_cnt += AYE.frac_clock;

	while (AYE.env_cnt >= AYE.env_div)
	{
		AYE.env_cnt -= AYE.env_div;

		//advance envelope logic implemented accordingly to the control bits

		AYE.env_amp += AYE.env_amp_delta;

		if (!(AYE.env_mode&AY_ENV_FLAGS_CONTINUE))
		{
			if ((AYE.env_amp < 0) || (AYE.env_amp > 15))
			{
				AYE.env_amp = 0;
				AYE.env_amp_delta = 0;
			}
		}
		else
		{
			if (AYE.env_mode&AY_ENV_FLAGS_HOLD)
			{
				if (AYE.env_amp < 0)
				{
					AYE.env_amp = 0;
					AYE.env_amp_delta = 0;
				}

				if (AYE.env_amp > 15)
				{
					AYE.env_amp = 15;
					AYE.env_amp_delta = 0;
				}
			}
			else
			{
				if (AYE.env_mode&AY_ENV_FLAGS_ALTERNATE)
				{
					if (AYE.env_amp < 0)
					{
						AYE.env_amp = 0;
						AYE.env_amp_delta = 1;
					}

					if (AYE.env_amp > 15)
					{
						AYE.env_amp = 15;
						AYE.env_amp_delta = -1;
					}
				}
				else
				{
					if (AYE.env_amp < 0) AYE.env_amp = 15;
					if (AYE.env_amp > 15) AYE.env_amp = 0;
				}
			}
		}
	}

	//mix outputs and set DACs

	uint8_t ta = ((AYE.mixer & 0x01) | (AYE.tone_a_out & 1)) ? 1 : 0;
	uint8_t tb = ((AYE.mixer & 0x02) | (AYE.tone_b_out & 1)) ? 1 : 0;
	uint8_t tc = ((AYE.mixer & 0x04) | (AYE.tone_c_out & 1)) ? 1 : 0;

	uint8_t na = ((AYE.mixer & 0x08) | AYE.noise_out) ? 1 : 0;
	uint8_t nb = ((AYE.mixer & 0x10) | AYE.noise_out) ? 1 : 0;
	uint8_t nc = ((AYE.mixer & 0x20) | AYE.noise_out) ? 1 : 0;

	if (ta&na) AYE.dac_a = (AYE.amp_a & 0x10) ? AYE.env_amp : AYE.amp_a; else AYE.dac_a = 0;
	if (tb&nb) AYE.dac_b = (AYE.amp_b & 0x10) ? AYE.env_amp : AYE.amp_b; else AYE.dac_b = 0;
	if (tc&nc) AYE.dac_c = (AYE.amp_c & 0x10) ? AYE.env_amp : AYE.amp_c; else AYE.dac_c = 0;
}



inline uint8_t AY_EmuGetDAC_A(void)
{
	return AYE.dac_a;
}



inline uint8_t AY_EmuGetDAC_B(void)
{
	return AYE.dac_b;
}



inline uint8_t AY_EmuGetDAC_C(void)
{
	return AYE.dac_c;
}
